-- HazLoot library containing the data for a classification list

HazLootList = {}
if not LibStub then -- not in WoW
  require("TheLoot")
  HLL = HazLootList
  HazLootList.standalone = true
end

HazLootList.ItemCount = 0
HazLootList.RoleByItem = {} -- Database of items
-- Why aren't these bitmaps? Because bitmaps aren't native to Lua, this is
-- just as (in)efficient as using the bitmap libraries, and probably easier
-- for laypeople to modify.
HazLootList.UNCLASSIFIED = 0
HazLootList.ROLES = { "Random roll",  -- 100
                      "Tier piece",   -- 200
                      "Caster DPS",   -- 300
                      "Healer",       -- 400
                      "Physical DPS", -- 500
                      "Tank",         -- 600
}
HazLootList.ROLES[0] = "<unclassified>"
HazLootList.ROLEVALUE = { ["<unclassified>"]=0,
                          ["Random roll"]=100,
                          ["Tier piece"]=200,
                          ["Caster DPS"]=300,
                          ["Healer"]=400,
                          ["Physical DPS"]=500,
                          ["Tank"]=600,
}
HazLootList.ANYRANDOM = HazLootList.ROLES["Random roll"]
HazLootList.RESTRICTIONS = { -- "Any" is a special case: 0
                         "Death Knight", --1
                         "Druid",        --2
                         "Hunter",       --3
                         "Mage",         --4
                         "Paladin",      --5
                         "Priest",       --6
                         "Rogue",        --7
                         "Shaman",       --8
                         "Warlock",      --9
                         "Warrior",      --10
}
HazLootList.RESTVALUES = {
                         ["Death Knight"]=1,
                         ["Druid"]=2,
                         ["Hunter"]=3,
                         ["Mage"]=4,
                         ["Paladin"]=5,
                         ["Priest"]=6,
                         ["Rogue"]=7,
                         ["Shaman"]=8,
                         ["Warlock"]=9,
                         ["Warrior"]=10,
}
HazLootList.RESTRICTIONS[0] = "<unclassified>"

-- itemRoleString:
-- itemName [ | role[: restriction] [, role[: restriction] ] ]
-- if no role is specified, the default, "Random roll" is used
-- e.g., the following are all valid:
--   Rod of Blasting
--   Rod of Lordly Might|Tank,Physical DPS:Paladin
--   Wand of Fireball|Caster DPS
--   Phylactery of Pain|Caster DPS,Healer
--   Codpiece of the Lost Wastrel|Tier piece:Hunter,Tier piece:Shaman,Tier piece:Warrior
-- the following are NOT valid and will be ignored:
--   Shield of This|Wall of That|Tank
--   Mace of Hitting,Physical DPS
--   Codpiece of the Lost Cad|Tier piece:Druid,Rogue,Paladin
function HazLootList:AddItem(itemRoleString)
  local sep = string.find(itemRoleString,"|")
  if nil == sep then -- Just an item name, any random
    -- print("Adding a bare item, any random: "..itemRoleString)
    HazLootList.RoleByItem[itemRoleString] = {HazLootList.ANYRANDOM}
    HazLootList.ItemCount = HazLootList.ItemCount + 1
    -- print(itemRoleString.." ADDED")
    return itemRoleString
  end
  -- print("loot role string has a | separator")
  local itemName  = string.sub(itemRoleString, 1, sep - 1)
  -- print("extracted item name: "..itemName)
  HazLootList.RoleByItem[itemName] = {}
  local roles = string.sub(itemRoleString, sep + 1 )
  -- print("extracted roles: "..roles)
  local restRole, restrict
  local i = 0
  local roleValue
  for roleName in string.gmatch(roles,"([^,]+),?") do
    i = i + 1
    if nil == roleName then
      return "nil - a problem"
    end
    -- is there a restriction?
    sep = string.find(roleName,":")
    if nil == sep then -- No
      -- print("role is unrestricted, adding role "..i.." to "..itemName..", with value "..HazLootList:GetRoleValue(roleName))
      HazLootList.RoleByItem[itemName][i]=HazLootList:GetRoleValue(roleName)
    else -- yes, what is it?
      restRole   = string.sub(roleName, 1, sep - 1)
      restrict   = string.sub(roleName, sep + 1 )
      HazLootList.RoleByItem[itemName][i] = HazLootList:GetRoleValue(restRole) + HazLootList:GetRestriction(restrict)
    end
  end
  -- print(itemRoleString.." ADDED")
  HazLootList.ItemCount = HazLootList.ItemCount + 1
  -- print("Now we have "..HazLootList.ItemCount.." items")
end

-- Add an "Unclassified" entry for this item
function HazLootList:AddUnclassifiedtem(itemName)
  HazLootList.RoleByItem[itemName] = {HazLootList.UNCLASSIFIED}
  HazLootList.ItemCount = HazLootList.ItemCount + 1
end

function HazLootList:GetRoleValue(roleName)
  local roleValue = HazLootList.ROLEVALUE[roleName]
  if nil == roleValue then -- invalid input
    return 0
  end
  return roleValue
end

function HazLootList:GetRestriction(restrict)
  local rValue = HazLootList.RESTVALUES[restrict]
  if nil == rValue then -- invalid input
    -- TODO: make this create a new classification?
    return "nil - a problem"
  end
  return rValue
end

-- Lazily parse the item string for the role table
function HazLootList:GetRoleTable(itemName)
  if not HazLootList.RoleByItem[itemName] then
    -- print(itemName.." is not loaded, getting the loot string")
    local lootString = TheLoot:GetLootString(itemName)
    if lootString then
      -- print("..found it. Adding the hammer's string:"..lootString)
      HazLootList:AddItem(lootString)
      -- print("HazLootList.RoleByItem["..itemName.."] should not now be nil, it should be a table. It is:"..  type(HazLootList.RoleByItem[itemName]))
    else
      -- print("got no loot string. Adding item as unclassified")
      HazLootList:AddUnclassifiedtem(itemName)
    end
  end
  return HazLootList.RoleByItem[itemName]
end

-- TODO: convert to itemId for localization
function HazLootList:GetRolesOf(itemName)
  HazLoot:Debug("finding role for "..itemName)
  local roleTable = HazLootList.RoleByItem[itemName]
  if not roleTable then roleTable = HazLootList:GetRoleTable(itemName) end
  local outString = ""
  if nil == roleTable then
    return HazLootList.ROLES[HazLootList.UNCLASSIFIED]
  end
  for i,role in ipairs(roleTable) do
    if i > 1 then outString = outString .. ", " end
    itemRole, itemRestriction = HazLootList:GetRoleName(role)
    outString = outString .. itemRole .. (nil == itemRestriction and "" or ":"..itemRestriction)
  end
  return outString
end

function HazLootList:IsTierPiece(itemId)
  local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType,
        itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(itemId)
  if nil == HazLootList.RoleByItem[itemName] then return false end
  local role = HazLootList.RoleByItem[itemName][1]
  return HazLootList.ROLEVALUE["Tier piece"] == (role - role%100)
  -- return HazLootList.ROLEVALUE["Tier piece"] == (role - math.mod(role,100))
end

function HazLootList:GetRoleName(role)
  local restriction = role%100
  -- local restriction = math.mod(role,100)
  if 0 == restriction then return HazLootList.ROLES[role/100]
  elseif restriction > (#HazLootList.RESTRICTIONS) then return HazLootList.RESTRICTIONS[0]
  end
  return HazLootList.ROLES[(role - restriction)/100], HazLootList.RESTRICTIONS[restriction]
end

function HazLootList:HowMuchLoot()
  return HazLootList.ItemCount
end

-- Sort Loot: return true if item1 should be sorted before item2
-- item : {itemId,itemName,itemLink,loot_slot}
function HazLootList:LootOrderComparator(item1,item2)
  if nil == item2 then return true end
  local value1 = HazLootList:GetLootValue(item1.itemId)
  local value2 = HazLootList:GetLootValue(item2.itemId)
  return (value1 == value2) and (item1.itemId < item2.itemId) or (value1 < value2)
end

function HazLootList:GetLootValue(itemId)
  -- 1. Tier pieces
  if HazLootList:IsTierPiece(itemId) then
    return 1 end
  local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType,
        itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(itemId)
  -- 2. Trinkets, Rings, Necks
  if itemEquipLoc == INVTYPE_NECK
     or itemEquipLoc == INVTYPE_FINGER
     or itemEquipLoc == INVTYPE_TRINKET
  then 
    return 2 end
  -- 3. Weapons, Shields, Wands, Relics, Idols
  if itemEquipLoc == INVTYPE_WEAPON
    or itemEquipLoc == INVTYPE_SHIELD
    or itemEquipLoc == INVTYPE_2HWEAPON
    or itemEquipLoc == INVTYPE_WEAPONMAINHAND
    or itemEquipLoc == INVTYPE_WEAPONOFFHAND
    or itemEquipLoc == INVTYPE_HOLDABLE
    or itemEquipLoc == INVTYPE_RANGED
    or itemEquipLoc == INVTYPE_THROWN
    or itemEquipLoc == INVTYPE_RANGEDRIGHT
    or itemEquipLoc == INVTYPE_RELIC
    or itemSubType  == "Miscellaneous"
  then
    return 3 end
  -- TODO: figure out how to localize this
  if itemType == "Armor" then
    if itemSubType == "Plate"
      then return 4 end
    if itemSubType == "Mail"
      then return 5 end
    if itemSubType == "Leather"
      then return 6 end
    if itemSubType == "Cloth"
      then return 7 end
  end
  return 0 -- unclassified
end

function HLL:Print(message)
  if HazLoot then
    HazLoot:Print(message)
  else
    print("[HLL]        "..message)
  end
end

function HLL:Debug(message)
  if HazLoot then
    HazLoot:Debug(message)
  else
    print("[HLL][DEBUG] "..message)
  end
end

-------- TESTING CODE --------
--[[
function HLLMockGetItemInfo(itemId)
  return TestItemNames[i], -- name
  "[itemLink"..i.."]",     -- link
  i,                       -- rarity
  100+i,                   -- level
  80,                      -- minlevel
  "Armor",                 -- type
  TestItemArmor[i],        -- subtype
  "1",                     -- stackcount
  INVTYPE_LEGS,            -- equiploc
  ""                       -- texture
end
TestItemArmor = {
"Cloth",
"Leather",
"Plate",
"Plate",
"Mail",
"Cloth",
}
TestItemNames = {
TestItemArmor[1].." Any Healer",
TestItemArmor[2].." Druid Healer",
TestItemArmor[3].." DK Physical DPS and Warrior tank",
TestItemArmor[4].." Random roll",
TestItemArmor[5].." Warrior Hunter Shaman tier piece",
TestItemArmor[6].." unclassified",
}
TestInput = {
TestItemNames[1].."|Healer",
TestItemNames[2].."|Healer:Druid",
TestItemNames[3].."|Physical DPS:Death Knight,Tank:Warrior",
TestItemNames[4].."|Random roll",
TestItemNames[5].."|Tier piece:Warrior,Tier piece:Hunter,Tier piece:Shaman",
TestItemNames[6].."|Some crap",
}
TestItems = {
  [ TestItemNames[1] ]={400},
  [ TestItemNames[2] ]={402},
  [ TestItemNames[3] ]={501,610},
  [ TestItemNames[4] ]={100},
  [ TestItemNames[5] ]={203,208,210},
  [ TestItemNames[6] ]={199},
}

function HazLootList:Test()
  local testItemName = "Smelting Pants"
  local tmp = TheLoot:GetLootString(testItemName)
  print("found loot string: "..tmp.." which should not yet be loaded")
  if assert(nil==HazLootList.RoleByItem[tmp],"...but it mysteriously is") then
    print("getting a loot string does not load the item")
  end
  print("now getting role table for "..testItemName.." which should load it")
  tmp = HazLootList:GetRoleTable(testItemName)
  if assert(nil ~= HazLootList.RoleByItem[testItemName],"...but it mysteriously hasn't") then
    print("loading the role table does load the item")
  end
  print("Now that it's loaded, check the roles")
  tmp = HazLootList:GetRolesOf(testItemName)
  print(" roles: "..tmp)
  -- Mock the GetItemInfo function to test the comparator
  -- add some fake loots
  MockLootBag = {}
  tlen = (#TestInput)
  for slot = 1,tlen do
    HazLootList:AddItem(TestInput[slot])
    print("Inserting "..slot..": "..TestItemNames[slot])
    table.insert(MockLootBag, {
      itemId = slot,
      itemName  = TestItemNames[slot],
      itemLink  = "["..TestItemNames[slot].."]",
      loot_slot = 1+tlen-slot,
    })
  end
  table.sort(MockLootBag, function(a,b) return HazLootList:LootOrderComparator(a,b) end)
  print("should now be sorted correctly:")
  for order,info in ipairs(MockLootBag) do
    print("order "..order..". itemId: "..info.itemId..", link: "..info.itemLink..", bag slot:"..info.loot_slot)
  end
end

HazLootList:Test()
function HLL:SetupMocks()
  HLL:Debug("Setting up mock methods")
  GetItemInfo = HLLMockGetItemInfo
  INVTYPE_LEGS = 1
end
if HLL.standalone then
  HLL:SetupMocks()
end
]]--
